#include "IZ_app.h"

IZ_ProcedureResult IZ_AppWSClientInitialize(IZ_App* app) {
	IZ_WSClientInitializeParams params = {
		.address = "localhost",
		.path = "/",
		.port = 42069,
	};

	if (IZ_WSClientInitialize(&app->client, params)) {
		printf("error\n");
		return -1;
	}
	return 0;
}

IZ_ProcedureResult IZ_AppConnect(void* app_raw) {
	IZ_App* app = app_raw;
	if (IZ_AppWSClientInitialize(app)) {
		return -1;
	}

	i32 result = 0;
	while (true) {
		if (IZ_WSClientHandle(&app->client)) {
			result = 1;
			break;
		}

		if (app->client.ws.interrupted) {
			break;
		}
	}

	IZ_WSClientTeardown(&app->client);
	return result;
}

void IZ_AppEstablishConnection(IZ_App* app) {
	app->client_thread = SDL_CreateThread(IZ_AppConnect, "networking", app);
	SDL_DetachThread(app->client_thread);
}

void IZ_AppCloseConnection(IZ_App* app) {
	if (!app->client.ws.context) {
		return;
	}
	IZ_WSClientCancelService(&app->client);
}

IZ_ProcedureResult IZ_AppInitialize(IZ_App* app, u8 argc, const char* argv[]) {
	memset(app, 0, sizeof(IZ_App));
	u32 flags = (
		SDL_INIT_VIDEO
		| SDL_INIT_GAMECONTROLLER
		| SDL_INIT_EVENTS
	);

	if (SDL_Init(flags) < 0) {
		SDL_LogError(SDL_LOG_CATEGORY_ERROR, "SDL could not initialize! SDL_Error: %s\n", SDL_GetError());
		return IZ_APP_RUN_SDL_INIT_ERROR;
	}

	const char* cmdline_buffer;
	char config_path[128];
	// TODO abstract command line args parsing
	if ((cmdline_buffer = lws_cmdline_option(argc, argv, "-c"))) {
		memcpy_s(config_path, 128, cmdline_buffer, 128);
	} else {
		IZ_ConfigGetDefaultPath(config_path, 128);
	}

	if (IZ_VideoInitialize(&app->video_state, config_path, argc, argv)) {
		return IZ_APP_RUN_VIDEO_INIT_ERROR;
	}

	if (IZ_InputInitialize(&app->input_state, config_path, argc, argv)) {
		return IZ_APP_RUN_INPUT_INIT_ERROR;
	}

	IZ_PoolInitialize(&app->pool, POOL_MAX_SIZE);

  // TODO put into its timer module
	app->ticks = 0;
	return IZ_APP_RUN_RESULT_OK;
}

void IZ_AppTeardown(IZ_App* app) {
	IZ_AppCloseConnection(app);
	IZ_PoolTeardown(&app->pool);
	IZ_InputTeardown(&app->input_state);
	IZ_VideoTeardown(&app->video_state);
	SDL_Quit();
}

IZ_ProcedureResult IZ_AppHandleSDLEvents(IZ_App* app) {
	SDL_Event e;
	while (SDL_PollEvent(&e) != 0) {
		if (e.type == SDL_QUIT) {
			return 1;
		}

		if (e.type == SDL_KEYDOWN) {
			if (e.key.keysym.sym == SDLK_PAGEUP && !app->client.ws.context) {
				IZ_AppEstablishConnection(app);
			} else if (e.key.keysym.sym == SDLK_PAGEDOWN) {
				IZ_AppCloseConnection(app);
			}
		}

		IZ_InputHandleSDLEvents(&app->input_state, e);
	}
	return 0;
}

void IZ_AppHandlePortMIDIEvents(IZ_App* app) {
	u8 player_index;
	i32* midi_events_count;
	u32 midi_event_index;
	for (player_index = 0; player_index < PLAYERS; player_index += 1) {
		if (!app->input_state.midi_input_state[player_index].device_info) {
			continue;
		}

		midi_events_count = &app->input_state.midi_input_state[player_index].midi_events_count;
		*midi_events_count = Pm_Read(
			app->input_state.midi_input_state[player_index].stream,
			// TODO bind buffers and streams to device instead of player input state
			app->input_state.midi_input_state[player_index].event_buffer,
			1024
		);

		if (*midi_events_count < 1) {
			continue;
		}

		for (midi_event_index = 0; midi_event_index < *midi_events_count; midi_event_index += 1) {
			IZ_InputHandlePortMIDIEvents(
				&app->input_state,
				app->input_state.midi_input_state[player_index].event_buffer[midi_event_index]
			);
		}
	}
}

IZ_ProcedureResult IZ_AppHandleEvents(IZ_App* app) {
	if (IZ_AppHandleSDLEvents(app)) {
		return 1;
	}
	IZ_AppHandlePortMIDIEvents(app);
	return 0;
}

IZ_ProcedureResult IZ_AppRunNetworkingThread(void* ptr) {
	return 0;
}

IZ_ProcedureResult IZ_AppRun(IZ_App* app, u8 argc, const char* argv[]) {
	IZ_ProcedureResult init_result = IZ_AppInitialize(app, argc, argv);
	if (init_result) {
		return init_result;
	}

//	if (IZ_WSClientConnect(&app->client, (IZ_WSClientConnectParams) {
//		.userdata = app,
//		.path = "/",
//		.protocol = NETWORK_PROTOCOL, // TODO handle ws protocol correctly
//		.address = "localhost",
//		.port = 42069,
//		.callback = IZ_AppWSClientCallback,
//	})) {
//		return IZ_APP_RUN_NETWORKING_ERROR;
//	}

	//app->client_thread = SDL_CreateThread(IZ_AppRunNetworkingThread, "Networking", app);
	//SDL_DetachThread(app->client_thread);

	while (true) {
		app->ticks = SDL_GetTicks64();

		// TODO do audio processing
		// TODO do networking
		//if (app->client.connection) {
			// FIXME stuck in infinite loop
			//IZ_WSClientHandle(&app->client);
		//}

		if (IZ_AppHandleEvents(app)) {
			break;
		}

		IZ_VideoUpdate(&app->video_state, app->ticks, &app->input_state);
	}

	IZ_AppTeardown(app);
	return IZ_APP_RUN_RESULT_OK;
}
