#include "IZ_wsserver.h"

void IZ_ProtocolCullLaggingClients(IZ_WSServerVHostData *vhd) {
	u32 oldest_tail = lws_ring_get_oldest_tail(vhd->ring);
	IZ_WSServerSessionData *old_pss = NULL;
	i32 most = 0;
	i32 before = (i32) lws_ring_get_count_waiting_elements(vhd->ring, &oldest_tail);
	i32 m;

	/*
	 * At least one guy with the oldest tail has lagged too far, filling
	 * the ringbuffer with stuff waiting for them, while new stuff is
	 * coming in, and they must close, freeing up ringbuffer entries.
	 */

	lws_start_foreach_llp_safe(
		IZ_WSServerSessionData**,
		ppss,
		vhd->pss_list,
		pss_list
	) {
		if ((*ppss)->tail == oldest_tail) {
			old_pss = *ppss;

			lwsl_user("Killing lagging client %p\n", (*ppss)->wsi);

			lws_set_timeout((*ppss)->wsi, PENDING_TIMEOUT_LAGGING,
				/*
				 * we may kill the wsi we came in on,
				 * so the actual close is deferred
				 */
											LWS_TO_KILL_ASYNC);

			/*
			 * We might try to write something before we get a
			 * chance to close.  But this pss is now detached
			 * from the ring buffer.  Mark this pss as culled so we
			 * don't try to do anything more with it.
			 */

			(*ppss)->culled = true;

			/*
			 * Because we can't kill it synchronously, but we
			 * know it's closing momentarily and don't want its
			 * participation any more, remove its pss from the
			 * vhd pss list early.  (This is safe to repeat
			 * uselessly later in the close flow).
			 *
			 * Notice this changes *ppss!
			 */

			lws_ll_fwd_remove(IZ_WSServerSessionData, pss_list, (*ppss), vhd->pss_list);

			/* use the changed *ppss so we won't skip anything */

			continue;
		}

		/*
		 * so this guy is a survivor of the cull.  Let's track
		 * what is the largest number of pending ring elements
		 * for any survivor.
		 */
		m = (i32) lws_ring_get_count_waiting_elements(vhd->ring, &((*ppss)->tail));
		if (m > most) {
			most = m;
		}
	} lws_end_foreach_llp_safe(ppss);

	/* it would mean we lost track of oldest... but Coverity insists */
	if (!old_pss) {
		return;
	}

	/*
	 * Let's recover (ie, free up) all the ring slots between the
	 * original oldest's last one and the "worst" survivor.
	 */

	lws_ring_consume_and_update_oldest_tail(
		vhd->ring,
		IZ_WSServerSessionData,
		&old_pss->tail,
		(size_t) (before - most),
		vhd->pss_list,
		tail,
		pss_list
	);

	lwsl_user("%s: shrunk ring from %d to %d\n", __func__, before, most);
}

/* destroys the message when everyone has had a copy of it */

IZ_ProcedureResult IZ_WSServerProtocolInitialize(struct lws* wsi) {
	IZ_WSServerVHostData* vhd_instance = (IZ_WSServerVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);
	IZ_WSServerVHostData** vhd = &vhd_instance;
	*vhd = lws_protocol_vh_priv_zalloc(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi),
		sizeof(IZ_WSServerVHostData)
	);
	(*vhd)->context = lws_get_context(wsi);
	(*vhd)->protocol = lws_get_protocol(wsi);
	(*vhd)->vhost = lws_get_vhost(wsi);
	(*vhd)->ring = lws_ring_create(
		sizeof(IZ_WebsocketMessage),
		RING_COUNT,
		IZ_WebsocketDestroyMessage
	);
	if (!(*vhd)->ring) {
		return 1;
	}
	return 0;
}

void IZ_WSServerProtocolTeardown(struct lws* wsi) {
	IZ_WSServerVHostData* vhd = (IZ_WSServerVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);
	lws_ring_destroy(vhd->ring);
}

void IZ_WSServerOnOpen(struct lws* wsi, IZ_WSServerSessionData* pss) {
	IZ_WSServerVHostData* vhd = (IZ_WSServerVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	/* add ourselves to the list of live pss held in the vhd */
	lwsl_user("LWS_CALLBACK_ESTABLISHED: wsi %p\n", wsi);
	lws_ll_fwd_insert(pss, pss_list, vhd->pss_list);
	pss->tail = lws_ring_get_oldest_tail(vhd->ring);
	pss->wsi = wsi;
}

void IZ_WSServerOnClose(struct lws* wsi, IZ_WSServerSessionData* pss) {
	IZ_WSServerVHostData* vhd = (IZ_WSServerVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	lwsl_user("LWS_CALLBACK_CLOSED: wsi %p\n", wsi);
	/* remove our closing pss from the list of live pss */
	lws_ll_fwd_remove(IZ_WSServerSessionData, pss_list, pss, vhd->pss_list);
}

IZ_ProcedureResult IZ_WSServerWritable(struct lws* wsi, IZ_WSServerSessionData* pss) {
	IZ_WSServerVHostData* vhd = (IZ_WSServerVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	if (pss->culled) {
		return 0;
	}
	const IZ_WebsocketMessage* pmsg = lws_ring_get_element(vhd->ring, &pss->tail);
	if (!pmsg) {
		return 0;
	}

	/* notice we allowed for LWS_PRE in the payload already */
	i32 m = lws_write(
		wsi,
		((unsigned char*) pmsg->payload) + LWS_PRE,
		pmsg->len,
		pmsg->binary ? LWS_WRITE_BINARY : LWS_WRITE_TEXT
	);

	if (m < (i32) pmsg->len) {
		lwsl_err("ERROR %d writing to ws socket\n", m);
		return -1;
	}

	lws_ring_consume_and_update_oldest_tail(
		vhd->ring,  /* lws_ring object */
		IZ_WSServerSessionData, /* type of objects with tails */
		&pss->tail,  /* tail of guy doing the consuming */
		1,    /* number of payload objects being consumed */
		vhd->pss_list,  /* head of list of objects with tails */
		tail,    /* member name of tail in objects with tails */
		pss_list  /* member name of next object in objects with tails */
	);

	/* more to do for us? */
	if (lws_ring_get_element(vhd->ring, &pss->tail)) {
		/* come back as soon as we can write more */
		lws_callback_on_writable(pss->wsi);
	}

	return 0;
}

void IZ_WSServerOnReceive(struct lws* wsi, void* in, size_t len) {
	IZ_WSServerVHostData* vhd = (IZ_WSServerVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);
	IZ_WebsocketMessage amsg;
	i32 n = (i32) lws_ring_get_count_free_elements(vhd->ring);
	if (!n) {
		/* forcibly make space */
		IZ_ProtocolCullLaggingClients(vhd);
		n = (i32) lws_ring_get_count_free_elements(vhd->ring);
	}

	if (!n) {
		return;
	}

	lwsl_user("LWS_CALLBACK_RECEIVE: free space %d\n", n);

	amsg.len = len;
	/* notice we over-allocate by LWS_PRE... */
	amsg.payload = malloc(LWS_PRE + len);
	amsg.binary = (u8) lws_frame_is_binary(wsi);
	if (!amsg.payload) {
		lwsl_user("OOM: dropping\n");
		return;
	}

	/* ...and we copy the payload in at +LWS_PRE */
	memcpy((char *) amsg.payload + LWS_PRE, in, len);
	if (!lws_ring_insert(vhd->ring, &amsg, 1)) {
		IZ_WebsocketDestroyMessage(&amsg);
		lwsl_user("dropping!\n");
		return;
	}

	/*
	 * let everybody know we want to write something on them
	 * as soon as they are ready
	 */
	lws_start_foreach_llp(IZ_WSServerSessionData**, ppss, vhd->pss_list) {
		lws_callback_on_writable((*ppss)->wsi);
	}

	lws_end_foreach_llp(ppss, pss_list);
}

IZ_ProcedureResult IZ_WSServerCallback(
	struct lws* wsi,
	enum lws_callback_reasons reason,
	void* user,
	void* in,
	size_t len
) {
	switch (reason) {
		case LWS_CALLBACK_PROTOCOL_INIT:
			return IZ_WSServerProtocolInitialize(wsi);
		case LWS_CALLBACK_PROTOCOL_DESTROY:
			IZ_WSServerProtocolTeardown(wsi);
			break;
		case LWS_CALLBACK_ESTABLISHED:
			IZ_WSServerOnOpen(wsi, user);
			break;
		case LWS_CALLBACK_CLOSED:
			IZ_WSServerOnClose(wsi, user);
			break;
		case LWS_CALLBACK_SERVER_WRITEABLE:
			return IZ_WSServerWritable(wsi, user);
		case LWS_CALLBACK_RECEIVE:
			IZ_WSServerOnReceive(wsi, in, len);
			break;
		default:
			break;
	}

	return 0;
}

IZ_ProcedureResult IZ_WSServerInitialize(IZ_WSServerState* state, IZ_WSServerInitializeParams params) {
	struct lws_context_creation_info info;
	memset(&info, 0, sizeof info);
	info.port = params.port;

	static struct lws_http_mount mount = {
		.mount_next = NULL,    /* linked-list "next" */
		.mountpoint = "/",    /* mountpoint URL */
		.origin = "./public", /* serve from dir */
		.def = "index.html",  /* default filename */
		.protocol = "http",
		.cgienv = NULL,
		.extra_mimetypes = NULL,
		.interpret = NULL,
		.cgi_timeout = 0,
		.cache_max_age = 0,
		.auth_mask = 0,
		.cache_reusable = 0,
		.cache_revalidate = 0,
		.cache_intermediaries = 0,
		.origin_protocol = LWSMPRO_FILE,  /* files in a dir */
		.mountpoint_len = 1,    /* char count */
		.basic_auth_login_file = NULL,
	};

	if (params.origin) {
		mount.origin = params.origin;
	}

	if (params.default_filename) {
		mount.def = params.default_filename;
	}

	if (params.protocol) {
		mount.protocol = params.protocol;
	}
	info.mounts = &mount;

	static const struct lws_protocols protocols[] = {
		{
			.name = NETWORK_PROTOCOL,
			.callback = IZ_WSServerCallback,
			.per_session_data_size = sizeof(IZ_WSServerSessionData),
			.rx_buffer_size = 0,
			.id = 0,
			.user = NULL,
			.tx_packet_size = 0,
		},
		{
			.name = "http",
			.callback = lws_callback_http_dummy,
			.per_session_data_size = 0,
			.rx_buffer_size = 0,
			.id = 0,
			.user = NULL,
			.tx_packet_size = 0,
		},
		LWS_PROTOCOL_LIST_TERM,
	};
	info.protocols = protocols;
	info.options = (
		LWS_SERVER_OPTION_VALIDATE_UTF8
		| LWS_SERVER_OPTION_HTTP_HEADERS_SECURITY_BEST_PRACTICES_ENFORCE
	);

	IZ_WebsocketInitialize(&state->ws);
	state->ws.context = lws_create_context(&info);
	if (!state->ws.context) {
		return -1;
	}

	return 0;
}

IZ_ProcedureResult IZ_WSServerHandle(IZ_WSServerState* state) {
	return IZ_WebsocketHandle(&state->ws);
}

void IZ_WSServerTeardown(IZ_WSServerState* state) {
	IZ_WebsocketTeardown(&state->ws);
}

void IZ_WSServerCancelService(IZ_WSServerState* state) {
	IZ_WebsocketCancelService(&state->ws);
}
